# views_admin.py
# Path: appointment/views_admin.py
# Url root: /app-admins/

"""
Author: Adams Pierre David
Since: 2.0.0
"""

import datetime
import json

from django.contrib import messages
from django.core.exceptions import ValidationError
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_POST

from appointment.decorators import (
    require_ajax, require_staff_or_superuser, require_superuser, require_user_authenticated)
from appointment.forms import PersonalInformationForm, ServiceForm, StaffAppointmentInformationForm, StaffMemberForm
from appointment.messages_ import appt_updated_successfully
from appointment.models import Appointment, DayOff, StaffMember, WorkingHours
from appointment.services import (
    create_new_appointment, create_staff_member_service, email_change_verification_service,
    fetch_user_appointments, handle_entity_management_request, handle_service_management_request,
    prepare_appointment_display_data, prepare_user_profile_data, save_appt_date_time, update_existing_appointment,
    update_personal_info_service)
from appointment.utils.db_helpers import (
    Service, get_day_off_by_id, get_staff_member_by_user_id, get_user_model,
    get_working_hours_by_id)
from appointment.utils.error_codes import ErrorCode
from appointment.utils.json_context import (
    convert_appointment_to_json, get_generic_context, get_generic_context_with_extra, handle_unauthorized_response,
    json_response)
from appointment.utils.permissions import check_extensive_permissions, check_permissions, \
    has_permission_to_delete_appointment


###############################################################


@require_user_authenticated
@require_staff_or_superuser
def get_user_appointments(request, response_type='html'):
    appointments = fetch_user_appointments(request.user)
    appointments_json = convert_appointment_to_json(request, appointments)

    if response_type == 'json':
        return json_response("Successfully fetched appointments.", custom_data={'appointments': appointments_json},
                             safe=False)

    # Render the HTML template
    extra_context = {
        'appointments': json.dumps(appointments_json),
    }
    context = get_generic_context_with_extra(request=request, extra=extra_context)
    # if appointment is empty and user doesn't have a staff-member instance, put a message
    # TODO: Refactor this logic, it's not clean
    if not appointments and not StaffMember.objects.filter(
            user=request.user).exists() and not request.user.is_superuser:
        messages.error(request, _("User doesn't have a staff member instance. Please contact the administrator."))
    return render(request, 'administration/staff_index.html', context)


@require_user_authenticated
@require_staff_or_superuser
def display_appointment(request, appointment_id):
    appointment, page_title, error_message, status_code = prepare_appointment_display_data(request.user, appointment_id)

    if error_message:
        context = get_generic_context(request=request)
        return render(request, 'error_pages/404_not_found.html', context=context, status=status_code)
    # If everything is okay, render the HTML template.
    extra_context = {
        'appointment': appointment,
        'page_title': page_title,
    }
    context = get_generic_context_with_extra(request=request, extra=extra_context)
    return render(request, 'administration/display_appointment.html', context)


@require_user_authenticated
@require_staff_or_superuser
def user_profile(request, staff_user_id=None):
    data = prepare_user_profile_data(request.user, staff_user_id)
    error = data.get('error')
    status_code = data.get('status_code', 400)
    context = get_generic_context_with_extra(request=request, extra=data['extra_context'])
    error_template = 'error_pages/403_forbidden.html' if status_code == 403 else 'error_pages/404_not_found.html'
    template = data['template'] if not error else error_template
    return render(request, template, context)


###############################################################


@require_user_authenticated
@require_staff_or_superuser
def add_day_off(request, staff_user_id=None, response_type='html'):
    staff_user_id = staff_user_id or request.user.pk
    if not check_permissions(staff_user_id, request.user):
        message = _("You can only add your own days off")
        return handle_unauthorized_response(request, message, response_type)

    staff_user_id = staff_user_id if staff_user_id else request.user.pk
    staff_member = get_staff_member_by_user_id(user_id=staff_user_id)
    return handle_entity_management_request(request, staff_member, entity_type='day_off')


@require_user_authenticated
@require_staff_or_superuser
def update_day_off(request, day_off_id, staff_user_id=None, response_type='html'):
    day_off = get_day_off_by_id(day_off_id)
    if not day_off:
        if response_type == 'json':
            return json_response("Day off does not exist.", status=404, success=False,
                                 error_code=ErrorCode.DAY_OFF_NOT_FOUND)
        else:
            context = get_generic_context(request=request)
            return render(request, 'error_pages/404_not_found.html', context=context, status=404)
    staff_user_id = staff_user_id or request.user.pk
    if not check_extensive_permissions(staff_user_id, request.user, day_off):
        message = _("You can only update your own days off.")
        return handle_unauthorized_response(request, message, response_type)
    staff_member = get_staff_member_by_user_id(user_id=staff_user_id)
    return handle_entity_management_request(request, staff_member, entity_type='day_off', instance=day_off)


@require_user_authenticated
@require_staff_or_superuser
def delete_day_off(request, day_off_id, staff_user_id=None):
    day_off = get_object_or_404(DayOff, pk=day_off_id)
    if not check_extensive_permissions(staff_user_id, request.user, day_off):
        message = _("You can only delete your own days off.")
        return handle_unauthorized_response(request, message, 'html')
    day_off.delete()
    if staff_user_id:
        return redirect('appointment:user_profile', staff_user_id=staff_user_id)
    return redirect('appointment:user_profile')


###############################################################


@require_user_authenticated
@require_staff_or_superuser
def add_working_hours(request, staff_user_id=None, response_type='html'):
    staff_user_id = staff_user_id or request.user.pk
    if not check_permissions(staff_user_id, request.user):
        message = _("You can only add your own working hours.")
        return handle_unauthorized_response(request, message, response_type)
    staff_user_id = staff_user_id if staff_user_id else request.user.pk
    staff_member = get_staff_member_by_user_id(user_id=staff_user_id)
    return handle_entity_management_request(request=request, staff_member=staff_member, staff_user_id=staff_user_id,
                                            entity_type='working_hours')


@require_user_authenticated
@require_staff_or_superuser
def update_working_hours(request, working_hours_id, staff_user_id=None, response_type='html'):
    working_hours = get_working_hours_by_id(working_hours_id)
    if not working_hours:
        if response_type == 'json':
            return json_response("Working hours does not exist.", status=404, success=False,
                                 error_code=ErrorCode.WORKING_HOURS_NOT_FOUND)
        else:
            context = get_generic_context(request=request)
            return render(request, 'error_pages/404_not_found.html', context=context)

    staff_user_id = staff_user_id or request.user.pk
    if not check_extensive_permissions(staff_user_id, request.user, working_hours):
        message = _("You can only update your own working hours.")
        return handle_unauthorized_response(request, message, response_type)
    staff_member = get_object_or_404(StaffMember, user_id=staff_user_id or request.user.id)
    return handle_entity_management_request(request=request, staff_member=staff_member, add=False,
                                            instance_id=working_hours_id, staff_user_id=staff_user_id,
                                            entity_type='working_hours', instance=working_hours)


@require_user_authenticated
@require_staff_or_superuser
def delete_working_hours(request, working_hours_id, staff_user_id=None):
    working_hours = get_object_or_404(WorkingHours, pk=working_hours_id)
    staff_member = working_hours.staff_member
    if not check_extensive_permissions(staff_user_id, request.user, working_hours):
        message = _("You can only delete your own working hours.")
        return handle_unauthorized_response(request, message, 'html')
    # update weekend hours if necessary
    staff_member.update_upon_working_hours_deletion(working_hours.day_of_week)

    working_hours.delete()

    if staff_user_id:
        return redirect('appointment:user_profile', staff_user_id=staff_user_id)
    return redirect('appointment:user_profile')


###############################################################


@require_user_authenticated
@require_staff_or_superuser
def add_or_update_staff_info(request, user_id=None):
    user = request.user

    # Only allow superusers or the authenticated user to edit his staff info
    if not check_permissions(staff_user_id=user_id, user=user):
        return json_response(_("Not authorized."), status=403, success=False, error_code=ErrorCode.NOT_AUTHORIZED)

    target_user = get_user_model().objects.get(pk=user_id) if user_id else user

    staff_member, created = StaffMember.objects.get_or_create(user=target_user)

    if request.method == 'POST':
        form = StaffAppointmentInformationForm(request.POST, instance=staff_member)
        if form.is_valid():
            form.save()
            if user_id:
                return redirect('appointment:user_profile', staff_user_id=user_id)
            return redirect('appointment:user_profile')
    else:
        form = StaffAppointmentInformationForm(instance=staff_member)

    context = get_generic_context_with_extra(request=request, extra={'form': form})
    return render(request, 'administration/manage_staff_member.html', context)


# TODO: Refactor this function, handle the different cases better.
@require_user_authenticated
@require_staff_or_superuser
def fetch_service_list_for_staff(request):
    appointment_id = request.GET.get('appointmentId')
    staff_id = request.GET.get('staff_member')
    if appointment_id:
        # Fetch services for a specific appointment (edit mode)
        if request.user.is_superuser:
            appointment = Appointment.objects.get(id=appointment_id)
            staff_member = appointment.get_staff_member()
        else:
            staff_member = StaffMember.objects.get(user=request.user)
            # Ensure the staff member is associated with this appointment
            if not Appointment.objects.filter(id=appointment_id,
                                              appointment_request__staff_member=staff_member).exists():
                return json_response(_("You do not have permission to access this appointment."), status_code=403)
        services = list(staff_member.get_services_offered().values('id', 'name'))
    elif staff_id:
        # Fetch services for the specified staff member (new mode based on staff member selection)
        staff_member = get_object_or_404(StaffMember, id=staff_id)
        services = list(staff_member.get_services_offered().values('id', 'name'))
    else:
        # Fetch all services for the staff member (create mode)
        try:
            staff_member = StaffMember.objects.get(user=request.user)
            services = list(staff_member.get_services_offered().values('id', 'name'))
        except StaffMember.DoesNotExist:
            if not request.user.is_superuser:
                return json_response(_("You're not a staff member. Can't perform this action !"), status=400,
                                     success=False)
            else:
                services = list(Service.objects.all().values('id', 'name'))

    if len(services) == 0:
        return json_response(_("No services offered by this staff member."), status=404, success=False,
                             error_code=ErrorCode.SERVICE_NOT_FOUND)
    return json_response(_("Successfully fetched services."), custom_data={'services_offered': services})


@require_user_authenticated
@require_superuser
def fetch_staff_list(request):
    staff_members = StaffMember.objects.all()
    staff_data = []
    for staff in staff_members:
        staff_data.append({
            'id': staff.id,
            'name': staff.get_staff_member_name(),
        })
    return json_response("Successfully fetched staff members.", custom_data={'staff_members': staff_data}, safe=False)


@require_user_authenticated
@require_staff_or_superuser
@require_ajax
@require_POST
def update_appt_min_info(request):
    data = json.loads(request.body)
    is_creating = data.get('isCreating', False)

    if is_creating:
        # Logic for creating a new appointment
        return create_new_appointment(data, request)
    else:
        # Logic for updating an existing appointment
        return update_existing_appointment(data, request)


@require_user_authenticated
@require_staff_or_superuser
@require_ajax
@require_POST
def validate_appointment_date(request):
    data = json.loads(request.body)
    start_time_str = data.get("start_time")
    appt_date_str = data.get("date")
    appointment_id = data.get("appointment_id")

    # Convert the string date and time to Python datetime objects
    start_time_obj = datetime.datetime.fromisoformat(start_time_str)
    appt_date = datetime.datetime.strptime(appt_date_str, "%Y-%m-%d").date()

    # Get the staff member for the appointment
    appt = Appointment.objects.get(id=appointment_id)
    staff_member = appt.appointment_request.staff_member

    # Check if the appointment's date and time are valid
    weekday: str = appt_date.strftime("%A")
    is_valid, message = Appointment.is_valid_date(appt_date, start_time_obj, staff_member, appointment_id, weekday)
    if not is_valid:
        return json_response(message, status=403, success=False, error_code=ErrorCode.INVALID_DATE)
    return json_response(_("Appointment date and time are valid."))


@require_user_authenticated
@require_staff_or_superuser
@require_ajax
@require_POST
def update_appt_date_time(request):
    data = json.loads(request.body)

    # Extracting the data
    start_time = data.get("start_time")
    appt_date = data.get("date")
    appointment_id = data.get("appointment_id")

    # save the data
    try:
        appt = save_appt_date_time(start_time, appt_date, appointment_id, request)
    except Appointment.DoesNotExist:
        return json_response(_("Appointment does not exist."), status=404, success=False,
                             error_code=ErrorCode.APPOINTMENT_NOT_FOUND)
    except ValidationError as e:
        return json_response(e.message, status=400, success=False)
    return json_response(appt_updated_successfully, custom_data={'appt': appt.id})


@require_user_authenticated
@require_staff_or_superuser
def update_personal_info(request, staff_user_id=None):
    # only superuser or the staff member itself can update the personal info
    if not check_permissions(staff_user_id=staff_user_id, user=request.user) or (
            not staff_user_id and not request.user.is_superuser):
        return json_response(_("Not authorized."), status=403, success=False, error_code=ErrorCode.NOT_AUTHORIZED)

    if request.method == 'POST':
        user, is_valid, error_message = update_personal_info_service(staff_user_id, request.POST, request.user)
        if is_valid:
            return redirect('appointment:user_profile')
        else:
            messages.error(request, error_message)
            return redirect('appointment:update_personal_info')

    if staff_user_id:
        user = get_user_model().objects.get(pk=staff_user_id)
    else:
        user = request.user

    form = PersonalInformationForm(initial={
        'first_name': user.first_name,
        'last_name': user.last_name,
        'email': user.email,
    }, user=user)

    context = get_generic_context_with_extra(request=request, extra={'form': form, 'btn_text': _("Update")})
    return render(request, 'administration/manage_staff_personal_info.html', context)


@require_user_authenticated
@require_staff_or_superuser
def email_change_verification_code(request):
    context = get_generic_context(request=request)

    if request.method == 'POST':
        code = request.POST.get('code')
        email = request.session.get('email')
        old_email = request.session.get('old_email')

        is_verified = email_change_verification_service(code, email, old_email)
        if is_verified:
            messages.success(request, _("Email updated successfully!"))
            return redirect('appointment:user_profile')
        else:
            messages.error(request, _("The verification code provided is incorrect. Please try again."))
            return render(request, 'administration/email_change_verification_code.html', context=context)

    return render(request, 'administration/email_change_verification_code.html', context=context)


###############################################################


@require_user_authenticated
@require_superuser
def add_staff_member_info(request):
    if request.method == 'POST':
        form = StaffMemberForm(request.POST)
        if form.is_valid():
            form.save()
            return redirect('appointment:user_profile')
    else:
        form = StaffMemberForm()

    context = get_generic_context_with_extra(request=request, extra={'form': form})
    return render(request, 'administration/manage_staff_member.html', context)


@require_user_authenticated
@require_superuser
def create_new_staff_member(request):
    if request.method == 'POST':
        user, is_valid, error_message = create_staff_member_service(request.POST, request)
        if is_valid:
            return redirect('appointment:user_profile', staff_user_id=user.pk)
        else:
            messages.error(request, error_message)
            return redirect('appointment:add_staff_member_personal_info')

    form = PersonalInformationForm()
    context = get_generic_context_with_extra(request=request, extra={'form': form, 'btn_text': _("Create")})
    return render(request, 'administration/manage_staff_personal_info.html', context=context)


@require_user_authenticated
@require_superuser
def make_superuser_staff_member(request):
    user = request.user
    StaffMember.objects.create(user=user)
    return redirect('appointment:user_profile')


@require_user_authenticated
@require_superuser
def remove_superuser_staff_member(request):
    user = request.user
    StaffMember.objects.filter(user=user).delete()
    return redirect('appointment:user_profile')


###############################################################
# Services

@require_user_authenticated
@require_superuser
def add_or_update_service(request, service_id=None, view=0):
    if request.method == 'POST':
        service, is_valid, error_message = handle_service_management_request(request.POST, request.FILES, service_id)
        if is_valid:
            messages.success(request, "Service saved successfully!")
            return redirect('appointment:add_service')
        else:
            messages.error(request, error_message)

    extra_context = {
        "btn_text": _("Save"),
        "page_title": _("Add Service"),
    }
    if service_id:
        service = get_object_or_404(Service, pk=service_id)
        form = ServiceForm(instance=service)
        if view != 1:
            extra_context['btn_text'] = _("Update")
            extra_context['page_title'] = _("Update Service")
        else:
            for field in form.fields.values():
                field.disabled = True
            extra_context['btn_text'] = None
            extra_context['page_title'] = _("View Service")
            extra_context['service'] = service
    else:
        form = ServiceForm()
    extra_context['form'] = form
    context = get_generic_context_with_extra(request=request, extra=extra_context)
    return render(request, 'administration/manage_service.html', context=context)


@require_user_authenticated
@require_superuser
def delete_service(request, service_id):
    service = get_object_or_404(Service, pk=service_id)
    service.delete()
    messages.success(request, _("Service deleted successfully!"))
    return redirect('appointment:user_profile')


###############################################################
# Remove staff member
@require_user_authenticated
@require_superuser
def remove_staff_member(request, staff_user_id):
    staff_member = get_object_or_404(StaffMember, user_id=staff_user_id)
    staff_member.delete()
    user = get_user_model().objects.get(pk=staff_user_id)
    if not user.is_superuser and user.is_staff:
        user.is_staff = False
        user.save()
    messages.success(request, _("Staff member deleted successfully!"))
    return redirect('appointment:user_profile')


@require_user_authenticated
@require_staff_or_superuser
def get_service_list(request, response_type='html'):
    services = Service.objects.all()
    if response_type == 'json':
        service_data = []
        for service in services:
            service_data.append({
                'id': service.id,
                'name': service.name,
                'description': service.description,
                'duration': service.get_duration(),
                'price': service.get_price_text(),
                'down_payment': service.get_down_payment_text(),
                'image': service.get_image_url(),
                'background_color': service.background_color
            })
        return json_response("Successfully fetched services.", custom_data={'services': service_data}, safe=False)
    context = get_generic_context_with_extra(request=request, extra={'services': services})
    return render(request, 'administration/service_list.html', context=context)


@require_user_authenticated
@require_staff_or_superuser
def delete_appointment(request, appointment_id):
    appointment = get_object_or_404(Appointment, pk=appointment_id)
    if not has_permission_to_delete_appointment(request.user, appointment):
        message = _("You can only delete your own appointments.")
        return handle_unauthorized_response(request, message, 'html')
    appointment.delete()
    messages.success(request, _("Appointment deleted successfully!"))
    return redirect('appointment:get_user_appointments')


@require_user_authenticated
@require_staff_or_superuser
def delete_appointment_ajax(request):
    data = json.loads(request.body)
    appointment_id = data.get("appointment_id")
    appointment = get_object_or_404(Appointment, pk=appointment_id)
    if not has_permission_to_delete_appointment(request.user, appointment):
        message = _("You can only delete your own appointments.")
        return json_response(message, status=403, success=False, error_code=ErrorCode.NOT_AUTHORIZED)
    appointment.delete()
    return json_response(_("Appointment deleted successfully."))


@require_user_authenticated
@require_staff_or_superuser
def is_user_staff_admin(request):
    user = request.user
    try:
        StaffMember.objects.get(user=user)
        return json_response(_("User is a staff member."), custom_data={'is_staff_admin': True})
    except StaffMember.DoesNotExist:
        # if superuser, all rights are granted even if not a staff member
        if not user.is_superuser:
            return json_response(_("User is not a staff member."), custom_data={'is_staff_admin': False})
        return json_response(_("User is a superuser."), custom_data={'is_staff_admin': True})



from datetime import time

from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.test import TestCase
from django.utils.translation import gettext as _

from appointment.models import WorkingHours
from appointment.tests.mixins.base_mixin import ServiceMixin, StaffMemberMixin, UserMixin


class WorkingHoursModelTestCase(TestCase, UserMixin, ServiceMixin, StaffMemberMixin):
    def setUp(self):
        self.user = self.create_user_()
        self.service = self.create_service_()
        self.staff_member = self.create_staff_member_(self.user, self.service)
        self.working_hours = WorkingHours.objects.create(
                staff_member=self.staff_member,
                day_of_week=1,
                start_time=time(9, 0),
                end_time=time(17, 0)
        )

    def test_default_attributes_on_creation(self):
        """Test if a WorkingHours instance can be created."""
        self.assertIsNotNone(self.working_hours)
        self.assertEqual(self.working_hours.staff_member, self.staff_member)
        self.assertEqual(self.working_hours.get_start_time(), time(9, 0))
        self.assertEqual(self.working_hours.get_end_time(), time(17, 0))

    def test_working_hours_str_method(self):
        """Test that the string representation of a WorkingHours instance is correct."""
        self.assertEqual(str(self.working_hours), "Monday - 09:00:00 to 17:00:00")

    def test_get_day_of_week_str(self):
        """Test that the get_day_of_week_str method in WorkingHours model works as expected."""
        self.assertEqual(self.working_hours.get_day_of_week_str(), _("Monday"))


class WorkingHoursValidationTestCase(TestCase, UserMixin, ServiceMixin, StaffMemberMixin):
    def setUp(self):
        self.user = self.create_user_()
        self.service = self.create_service_()
        self.staff_member = self.create_staff_member_(self.user, self.service)
        self.working_hours = WorkingHours.objects.create(
                staff_member=self.staff_member,
                day_of_week=1,
                start_time=time(9, 0),
                end_time=time(17, 0)
        )

    def test_working_hours_start_time_before_end_time(self):
        """A WorkingHours instance cannot be created if start_time is after end_time."""
        with self.assertRaises(ValidationError):
            WorkingHours.objects.create(
                staff_member=self.staff_member,
                day_of_week=2,
                start_time=time(17, 0),
                end_time=time(9, 0)
            ).clean()

    def test_working_hours_without_staff_member(self):
        """A WorkingHours instance cannot be created without a staff member."""
        with self.assertRaises(IntegrityError):
            WorkingHours.objects.create(
                day_of_week=3,
                start_time=time(9, 0),
                end_time=time(17, 0)
            )

    def test_working_hours_is_owner(self):
        """Test that is_owner method in WorkingHours model works as expected."""
        self.assertTrue(self.working_hours.is_owner(self.user.id))
        self.assertFalse(self.working_hours.is_owner(9999))  # Assuming 9999 is not a valid user ID in your tests

    def test_staff_member_weekend_status_update(self):
        """Test that the staff member's weekend status is updated when a WorkingHours instance is created."""
        WorkingHours.objects.create(
            staff_member=self.staff_member,
            day_of_week=6,  # Saturday
            start_time=time(9, 0),
            end_time=time(12, 0)
        )
        self.staff_member.refresh_from_db()
        self.assertTrue(self.staff_member.work_on_saturday)

        WorkingHours.objects.create(
            staff_member=self.staff_member,
            day_of_week=0,  # Sunday
            start_time=time(9, 0),
            end_time=time(12, 0)
        )
        self.staff_member.refresh_from_db()
        self.assertTrue(self.staff_member.work_on_sunday)

    def test_working_hours_duplicate_day(self):
        """A WorkingHours instance cannot be created if the staff member already has a working hours on that day."""
        with self.assertRaises(IntegrityError):
            WorkingHours.objects.create(
                staff_member=self.staff_member,
                day_of_week=1,  # Same day as the working_hours created in setUp
                start_time=time(9, 0),
                end_time=time(17, 0)
            )



# date_time.py
# Path: appointment/utils/date_time.py

"""
Author: Adams Pierre David
Since: 2.0.0
"""

import datetime

from django.utils import timezone
from django.utils.translation import gettext_lazy as _, ngettext


def combine_date_and_time(date, time) -> datetime.datetime:
    """Combine a date and a time into a datetime object.

    :param date: The date.
    :param time: The time.
    :return: A datetime object.
    """
    return datetime.datetime.combine(date, time)


def convert_12_hour_time_to_24_hour_time(time_to_convert) -> str:
    """Convert a 12-hour time to a 24-hour time.

    :param time_to_convert: The time to convert.
    :return: The converted time.
    :raises ValueError: If the input time is not in the correct format or is invalid.
    """
    if isinstance(time_to_convert, (datetime.datetime, datetime.time)):
        return time_to_convert.strftime('%H:%M:%S')
    elif isinstance(time_to_convert, str):
        try:
            time_str = time_to_convert.strip().upper()
            return datetime.datetime.strptime(time_str, '%I:%M %p').strftime('%H:%M:%S')
        except ValueError:
            raise ValueError(f"Invalid 12-hour time format: {time_to_convert}")
    else:
        raise ValueError(f"Unsupported data type for time conversion: {type(time_to_convert)}")


def convert_24_hour_time_to_12_hour_time(time_to_convert) -> str:
    """Convert a 24-hour time to a 12-hour time.

    :param time_to_convert: The time to convert in 'HH:MM' or 'HH:MM:SS' format, or a datetime.time object.
    :return: The converted time in 'HH:MM AM/PM' or 'HH:MM:SS AM/PM' format.
    :raises ValueError: If the input time is not in the correct format or is invalid.
    """
    # Handle datetime.time object directly
    if isinstance(time_to_convert, datetime.time):
        return time_to_convert.strftime('%I:%M %p')

    # Handle string input
    for source_fmt, dest_fmt in zip(['%H:%M:%S', '%H:%M'], ['%I:%M:%S %p', '%I:%M %p']):
        try:
            # Parse the input string according to the 24-hour format
            parsed_time = datetime.datetime.strptime(time_to_convert, source_fmt)
            # Convert and return the time in 12-hour format
            return parsed_time.strftime(dest_fmt)
        except ValueError:
            continue  # Try the next format if there was a parsing error

    # If input was not datetime.time and did not match string formats, raise an error
    raise ValueError(f"Invalid 24-hour time format: {time_to_convert}")


def convert_minutes_in_human_readable_format(minutes: float) -> str:
    """Convert a number of minutes in a human-readable format.

    :param minutes: The number of minutes to convert.
    :return: The converted minutes in a human-readable format.
    """
    if minutes == 0:
        return _("Not set.")
    if minutes < 0:
        raise ValueError("Minutes cannot be negative.")
    days, remaining_minutes = divmod(int(minutes), 1440)
    hours, minutes = divmod(int(remaining_minutes), 60)

    parts = []
    if days:
        days_display = ngettext("%(count)d day", "%(count)d days", days) % {'count': days}
        parts.append(days_display)

    if hours:
        hours_display = ngettext("%(count)d hour", "%(count)d hours", hours) % {'count': hours}
        parts.append(hours_display)

    if minutes:
        minutes_display = ngettext("%(count)d minute", "%(count)d minutes", minutes) % {'count': minutes}
        parts.append(minutes_display)

    if len(parts) == 1:
        return parts[0]
    elif len(parts) == 2:
        return _("{first_part} and {second_part}").format(first_part=parts[0], second_part=parts[1])
    elif len(parts) == 3:
        return _("{days}, {hours} and {minutes}").format(days=parts[0], hours=parts[1], minutes=parts[2])


def convert_str_to_date(date_str: str) -> datetime.date:
    """Convert a date string to a datetime date object.

    :param date_str: The date string.
                     Supported formats include `%Y-%m-%d` (like "2023-12-31") and `%Y/%m/%d` (like "2023/12/31").
    :return: The converted `datetime.date`'s object.
    """
    date_formats = ['%Y-%m-%d', '%Y/%m/%d', '%Y.%m.%d']

    for fmt in date_formats:
        try:
            return datetime.datetime.strptime(date_str, fmt).date()
        except ValueError:
            pass

    raise ValueError(f"Invalid date format for '{date_str}'. Supported formats are YYYY-MM-DD and YYYY/MM/DD.")


def convert_str_to_time(time_str: str) -> datetime.time:
    """Convert a string representation of time to a Python `time` object.

    The function tries both 12-hour and 24-hour formats.

    :param time_str: A string representation of time.
    :return: A Python `time` object.
    """
    formats = ["%I:%M %p", "%H:%M:%S", "%H:%M"]

    for fmt in formats:
        try:
            return datetime.datetime.strptime(time_str.strip().upper(), fmt).time()
        except ValueError:
            pass

    raise ValueError(
        f"Invalid time format for '{time_str}'. Expected either a 12-hour (e.g., '10:00 AM') or 24-hour (e.g., "
        f"'13:00:00') format.")


def get_ar_end_time(start_time, duration) -> datetime.time:
    """Get the end time of an appointment request based on the start time and the duration.

    :param start_time: The start time of the appointment request.
    :param duration: The duration in minutes or as timedelta of the appointment request.
    :return: The end time of the appointment request.
    """
    # Check types
    if not isinstance(start_time, (datetime.time, str)):
        raise TypeError("start_time must be a datetime.time object or a string in 'HH:MM:SS' format.")

    if not isinstance(duration, (datetime.timedelta, int, float)):
        raise TypeError("duration must be either a datetime.timedelta or a numeric type representing minutes.")

    if isinstance(duration, (int, float)) and duration < 0:
        raise ValueError("duration cannot be negative.")

    # Convert the time object to a datetime object
    if isinstance(start_time, str):
        start_time = convert_str_to_time(start_time)

    dt_start_time = datetime.datetime.combine(datetime.datetime.today(), start_time)

    # Convert duration to minutes if it's a timedelta
    if isinstance(duration, datetime.timedelta):
        duration_minutes = duration.total_seconds() / 60
    else:
        duration_minutes = int(duration)

    # Add the duration
    dt_end_time = dt_start_time + datetime.timedelta(minutes=duration_minutes)

    # If end time goes past midnight, wrap it around
    if dt_end_time.day > dt_start_time.day:
        dt_end_time = dt_end_time - datetime.timedelta(days=1)

    return dt_end_time.time()


def get_timestamp() -> str:
    """Get the current timestamp as a string without the decimal part.

    :return: The current timestamp (e.g. "1612345678")
    """
    timestamp = str(timezone.now().timestamp())
    return timestamp.replace('.', '')


def get_current_year() -> int:
    """Get the current year as an integer.

    :return: The current year
    """
    return datetime.datetime.now().year


def get_weekday_num(weekday: str) -> int:
    """Get the number of the weekday.

    :param weekday: The weekday (e.g. "Monday", "Tuesday", etc.)
    :return: The number of the weekday (0 for Sunday, 1 for Monday, etc.)
    """
    weekdays = {
        'monday': 1,
        'tuesday': 2,
        'wednesday': 3,
        'thursday': 4,
        'friday': 5,
        'saturday': 6,
        'sunday': 0
    }
    return weekdays.get(weekday.lower(), -1)


def time_difference(time1, time2):
    # If inputs are datetime.time objects, convert them to datetime.datetime objects for the same day
    if isinstance(time1, datetime.time) and isinstance(time2, datetime.time):
        today = datetime.datetime.today()
        datetime1 = datetime.datetime.combine(today, time1)
        datetime2 = datetime.datetime.combine(today, time2)
    elif isinstance(time1, datetime.datetime) and isinstance(time2, datetime.datetime):
        datetime1 = time1
        datetime2 = time2
    else:
        raise ValueError("Both inputs should be of the same type, either datetime.time or datetime.datetime")

    # Check if datetime2 is earlier than datetime1
    if datetime2 < datetime1:
        raise ValueError("The second time provided (time2) should not be earlier than the first time (time1).")

    # Find the difference
    delta = datetime2 - datetime1

    return delta



from copy import deepcopy
from datetime import datetime
from unittest import mock
from unittest.mock import MagicMock, patch

from django.test.client import RequestFactory
from django.utils import timezone
from django.utils.translation import gettext as _

from appointment.messages_ import thank_you_no_payment, thank_you_payment, thank_you_payment_plus_down
from appointment.models import AppointmentRescheduleHistory
from appointment.tests.base.base_test import BaseTest
from appointment.utils.email_ops import (
    get_thank_you_message, notify_admin_about_appointment, notify_admin_about_reschedule,
    send_reschedule_confirmation_email,
    send_reset_link_to_staff_member, send_thank_you_email,
    send_verification_email
)


class SendResetLinkToStaffMemberTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.user = deepcopy(self.users['staff1'])
        self.user.is_staff = True
        self.user.save()
        self.factory = RequestFactory()
        self.request = self.factory.get('/')
        self.email = 'daniel.jackson@django-appointment.com'

    @mock.patch('appointment.utils.email_ops.send_email')
    @mock.patch('appointment.models.PasswordResetToken.create_token')
    @mock.patch('appointment.utils.email_ops.get_absolute_url_')
    @mock.patch('appointment.utils.email_ops.get_website_name')
    def test_send_reset_link(self, mock_get_website_name, mock_get_absolute_url, mock_create_token, mock_send_email):
        # Set up the token
        mock_token = mock.Mock()
        mock_token.token = "Colonel_Samantha_Carter_a_Tau_ri_Scientist"
        mock_create_token.return_value = mock_token

        mock_get_absolute_url.return_value = f"http://gateroomserver/reset_password/{mock_token.token}"
        mock_get_website_name.return_value = 'Gate Room Server'

        send_reset_link_to_staff_member(self.user, self.request, self.email)

        # Check send_email was called with correct parameters
        mock_send_email.assert_called_once()
        args, kwargs = mock_send_email.call_args
        self.assertEqual(kwargs['recipient_list'], [self.email])
        self.assertIn('Gate Room Server', kwargs['message'])
        self.assertIn('http://gateroomserver/reset_password', kwargs['message'])
        self.assertIn('Colonel_Samantha_Carter_a_Tau_ri_Scientist', kwargs['message'])

        # Additional assertions to verify more parts of the message content
        self.assertIn('Hello', kwargs['message'])
        self.assertIn(self.user.first_name, kwargs['message'])
        self.assertIn(str(datetime.now().year), kwargs['message'])
        self.assertIn('No additional details provided.', kwargs['message'])
        self.assertIn(self.user.username, kwargs['message'])


class GetThankYouMessageTests(BaseTest):

    def setUp(self):
        super().setUp()
        self.ar = MagicMock()

    def test_thank_you_no_payment(self):
        with patch('appointment.utils.email_ops.APPOINTMENT_PAYMENT_URL', None):
            message = get_thank_you_message(self.ar)
            self.assertIn(thank_you_no_payment, message)

    def test_thank_you_payment_plus_down(self):
        with patch('appointment.utils.email_ops.APPOINTMENT_PAYMENT_URL', "http://payment.url"):
            self.ar.accepts_down_payment.return_value = True
            message = get_thank_you_message(self.ar)
            self.assertIn(thank_you_payment_plus_down, message)

    def test_thank_you_payment(self):
        with patch('appointment.utils.email_ops.APPOINTMENT_PAYMENT_URL', "http://payment.url"):
            self.ar.accepts_down_payment.return_value = False
            message = get_thank_you_message(self.ar)
            self.assertIn(thank_you_payment, message)


class SendThankYouEmailTests(BaseTest):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()
        self.request = self.factory.get('/')
        self.ar = self.create_appt_request_for_sm1()
        self.email = "georges.s.hammond@django-appointment.com"
        self.appointment_details = "Details about the appointment"
        self.account_details = "Details about the account"

    @patch('appointment.utils.email_ops.send_email')
    @patch('appointment.utils.email_ops.get_thank_you_message')
    def test_send_thank_you_email(self, mock_get_thank_you_message, mock_send_email):
        mock_get_thank_you_message.return_value = "Thank you message"

        send_thank_you_email(self.ar, self.users['client1'], self.request, self.email, self.appointment_details,
                             self.account_details)

        mock_send_email.assert_called_once()
        args, kwargs = mock_send_email.call_args
        self.assertIn(self.email, kwargs['recipient_list'])
        self.assertIn("Thank you message", kwargs['context']['message_1'])


class NotifyAdminAboutAppointmentTests(BaseTest):
    def setUp(self):
        super().setUp()
        self.appointment = self.create_appt_for_sm1()
        self.client_name = "Oma Desala"

    @patch('appointment.utils.email_ops.notify_admin')
    @patch('appointment.utils.email_ops.send_email')
    def test_notify_admin_about_appointment(self, mock_send_email, mock_notify_admin):
        notify_admin_about_appointment(self.appointment, self.client_name)
        mock_notify_admin.assert_called_once()
        mock_send_email.assert_called_once()


class SendVerificationEmailTests(BaseTest):
    def setUp(self):
        super().setUp()
        self.appointment = self.create_appt_for_sm1()
        self.email = "richard.woolsey@django-appointment.com"

    @patch('appointment.utils.email_ops.send_email')
    @patch('appointment.models.EmailVerificationCode.generate_code', return_value="123456")
    def test_send_verification_email(self, mock_generate_code, mock_send_email):
        user = MagicMock()

        send_verification_email(user, self.email)

        mock_send_email.assert_called_once_with(
                recipient_list=[self.email],
                subject=_("Email Verification"),
                message=mock.ANY
        )
        self.assertIn("123456", mock_send_email.call_args[1]['message'])


class SendRescheduleConfirmationEmailTests(BaseTest):
    def setUp(self):
        super().setUp()
        self.appointment_request = self.create_appt_request_for_sm1()
        self.reschedule_history = AppointmentRescheduleHistory.objects.create(
                appointment_request=self.appointment_request,
                date=self.appointment_request.date + timezone.timedelta(days=1),
                start_time=self.appointment_request.start_time,
                end_time=self.appointment_request.end_time,
                staff_member=self.staff_member1,
                reason_for_rescheduling="Had to reschedule because I got stuck in a time loop. Again"
        )
        self.first_name = "Jack"
        self.email = "jack.oneill@django-appointment.com"

    @mock.patch('appointment.utils.email_ops.get_absolute_url_')
    @mock.patch('appointment.utils.email_ops.send_email')
    def test_send_reschedule_confirmation_email(self, mock_send_email, mock_get_absolute_url):
        request = mock.MagicMock()
        mock_get_absolute_url.return_value = "http://gateroomserver/confirmation_link"

        send_reschedule_confirmation_email(request, self.reschedule_history, self.appointment_request, self.first_name,
                                           self.email)

        # Check if `send_email` was called correctly
        mock_send_email.assert_called_once()
        call_args, call_kwargs = mock_send_email.call_args

        self.assertEqual(call_kwargs['recipient_list'], [self.email])
        self.assertEqual(call_kwargs['subject'], _("Confirm Your Appointment Rescheduling"))
        self.assertIn('reschedule_date', call_kwargs['context'])
        self.assertIn('confirmation_link', call_kwargs['context'])
        self.assertEqual(call_kwargs['context']['confirmation_link'], "http://gateroomserver/confirmation_link")


class NotifyAdminAboutRescheduleTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.appointment_request = self.create_appt_request_for_sm1()
        self.reschedule_history = AppointmentRescheduleHistory.objects.create(
                appointment_request=self.appointment_request,
                date=self.appointment_request.date + timezone.timedelta(days=1),
                start_time=self.appointment_request.start_time,
                end_time=self.appointment_request.end_time,
                staff_member=self.staff_member1,
                reason_for_rescheduling="Captured by Anubis"
        )
        self.client_name = "Jonas Quinn"

    @patch('appointment.utils.email_ops.notify_admin')
    @patch('appointment.utils.email_ops.send_email')
    @patch('appointment.utils.email_ops.get_website_name', return_value="Stargate Command")
    @patch('appointment.utils.email_ops.convert_24_hour_time_to_12_hour_time',
           side_effect=lambda x: x.strftime("%I:%M %p"))
    def test_notify_admin_about_reschedule(self, mock_convert_time, mock_get_website_name, mock_send_email,
                                           mock_notify_admin):
        notify_admin_about_reschedule(self.reschedule_history, self.appointment_request, self.client_name)

        # Check if notify_admin was called correctly
        mock_notify_admin.assert_called_once()
        notify_admin_args, notify_admin_kwargs = mock_notify_admin.call_args
        self.assertIn(self.client_name, notify_admin_kwargs['subject'])
        self.assertEqual(notify_admin_kwargs['context']['client_name'], self.client_name)
        self.assertEqual(notify_admin_kwargs['context']['service_name'], self.appointment_request.service.name)
        self.assertEqual(notify_admin_kwargs['context']['reason_for_rescheduling'],
                         self.reschedule_history.reason_for_rescheduling)
        self.assertEqual(notify_admin_kwargs['context']['old_date'],
                         self.appointment_request.date.strftime("%A, %d %B %Y"))
        self.assertEqual(notify_admin_kwargs['context']['reschedule_date'],
                         self.reschedule_history.date.strftime("%A, %d %B %Y"))
        self.assertEqual(notify_admin_kwargs['context']['company'], "Stargate Command")



from datetime import date, time, timedelta

from appointment.models import (
    Appointment, AppointmentRequest, AppointmentRescheduleHistory, Config, Service, StaffMember
)
from appointment.utils.db_helpers import get_user_model


class UserMixin:
    def __init__(self):
        pass

    @classmethod
    def create_user_(cls, first_name="Janet", last_name="Fraiser", email="janet.fraiser@django-appointment.com",
                     username="janet.fraiser",
                     password="G0a'uld$Emp1re"):
        return get_user_model().objects.create_user(
                first_name=first_name,
                last_name=last_name,
                email=email,
                username=username,
                password=password
        )


class ServiceMixin:
    def __init__(self):
        pass

    @classmethod
    def create_service_(cls, name="Quantum Mirror Assessment", duration=timedelta(hours=1), price=50000,
                        description="Assess the Quantum Mirror"):
        return Service.objects.create(
                name=name,
                duration=duration,
                price=price,
                description=description
        )


class StaffMemberMixin:
    def __init__(self):
        pass

    @classmethod
    def create_staff_member_(cls, user, service):
        staff_member = StaffMember.objects.create(user=user)
        staff_member.services_offered.add(service)
        return staff_member


class AppointmentRequestMixin:
    def __init__(self):
        pass

    @classmethod
    def create_appointment_request_(cls, service, staff_member, date_=date.today(), start_time=time(9, 0),
                                    end_time=time(10, 0)):
        return AppointmentRequest.objects.create(
                date=date_,
                start_time=start_time,
                end_time=end_time,
                service=service,
                staff_member=staff_member
        )

    @classmethod
    def clean_appt_request_for_user_(cls, user):
        AppointmentRequest.objects.filter(staff_member__user=user).delete()


class AppointmentMixin:
    def __init__(self):
        pass

    @classmethod
    def create_appointment_(cls, user, appointment_request, phone="+12392340543",
                            address="Stargate Command, Cheyenne Mountain Complex, Colorado Springs, CO"):
        return Appointment.objects.create(
                client=user,
                appointment_request=appointment_request,
                phone=phone,
                address=address
        )

    @classmethod
    def clean_appointment_for_user_(cls, user):
        Appointment.objects.filter(client=user).delete()


class AppointmentRescheduleHistoryMixin:
    def __init__(self):
        pass

    @classmethod
    def create_reschedule_history_(cls, appointment_request, date_, start_time, end_time, staff_member,
                                   reason_for_rescheduling="Zat'nik'tel Discharge"):
        return AppointmentRescheduleHistory.objects.create(
                appointment_request=appointment_request,
                date=date_,
                start_time=start_time,
                end_time=end_time,
                staff_member=staff_member,
                reason_for_rescheduling=reason_for_rescheduling
        )


class ConfigMixin:
    def __init__(self):
        pass

    @classmethod
    def create_config_(cls, lead_time=time(9, 0), finish_time=time(17, 0), slot_duration=30,
                       appointment_buffer_time=0):
        return Config.objects.create(
                lead_time=lead_time,
                finish_time=finish_time,
                slot_duration=slot_duration,
                appointment_buffer_time=appointment_buffer_time
        )



# test_permissions.py
# Path: appointment/tests/utils/test_permissions.py

import datetime
from unittest import mock

from appointment.tests.base.base_test import BaseTest
from appointment.utils.db_helpers import WorkingHours
from appointment.utils.permissions import (
    check_entity_ownership, check_extensive_permissions, check_permissions, has_permission_to_delete_appointment
)


class CheckEntityOwnershipTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.superuser = self.users['superuser']
        self.superuser.is_superuser = True
        self.superuser.save()
        self.user1 = self.users['staff1']
        self.user2 = self.users['staff2']
        self.entity_owned_by_user1 = WorkingHours.objects.create(
                staff_member=self.staff_member1, day_of_week=0, start_time=datetime.time(8, 0),
                end_time=datetime.time(12, 0))

    def test_check_entity_ownership(self):
        """Test if ownership of an entity can be checked."""
        # User is the owner
        self.assertTrue(check_entity_ownership(self.user1, self.entity_owned_by_user1))
        # Superuser but not owner
        self.assertTrue(check_entity_ownership(self.superuser, self.entity_owned_by_user1))
        # Neither owner nor superuser
        self.assertFalse(check_entity_ownership(self.user2, self.entity_owned_by_user1))


class CheckExtensivePermissionsTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.superuser = self.users['superuser']
        self.superuser.is_superuser = True
        self.superuser.save()
        self.user1 = self.users['staff1']
        self.user2 = self.users['staff2']
        self.entity_owned_by_user1 = WorkingHours.objects.create(
                staff_member=self.staff_member1, day_of_week=0, start_time=datetime.time(8, 0),
                end_time=datetime.time(12, 0))

    def test_check_extensive_permissions(self):
        """Test if extensive permissions can be checked."""
        # staff_user_id matches and user owns entity
        self.assertTrue(check_extensive_permissions(self.user1.pk, self.user1, self.entity_owned_by_user1))
        # staff_user_id matches but user doesn't own entity
        self.assertFalse(check_extensive_permissions(self.user2.pk, self.user2, self.entity_owned_by_user1))
        # staff_user_id doesn't match but user is superuser
        self.assertTrue(check_extensive_permissions(None, self.superuser, self.entity_owned_by_user1))
        # staff_user_id matches and no entity provided
        self.assertTrue(check_extensive_permissions(self.user1.pk, self.user1, None))
        # Neither staff_user_id matches nor superuser
        self.assertFalse(check_extensive_permissions(None, self.user2, self.entity_owned_by_user1))


class CheckPermissionsTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.superuser = self.users['superuser']
        self.superuser.is_superuser = True
        self.superuser.save()
        self.user1 = self.users['staff1']
        self.user2 = self.users['staff2']

    def test_check_permissions(self):
        """Test if permissions can be checked."""
        # staff_user_id matches
        self.assertTrue(check_permissions(self.user1.pk, self.user1))
        # staff_user_id doesn't match but user is superuser
        self.assertTrue(check_permissions(None, self.superuser))
        # Neither staff_user_id matches nor superuser
        self.assertFalse(check_permissions(None, self.user2))


class HasPermissionToDeleteAppointmentTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.superuser = self.users['superuser']
        self.superuser.is_superuser = True
        self.superuser.save()
        self.user1 = self.users['staff1']
        self.user2 = self.users['staff2']
        self.appointment = self.create_appt_for_sm1()

    def test_has_permission_to_delete_appointment(self):
        """Test if the user has permission to delete the given appointment."""
        # Mock get_staff_member to return the staff member associated with the appointment
        with mock.patch('appointment.models.Appointment.get_staff_member', return_value=self.staff_member1):
            # User is the staff member associated with the appointment
            self.assertTrue(has_permission_to_delete_appointment(self.user1, self.appointment))
            # User is a superuser
            self.assertTrue(has_permission_to_delete_appointment(self.superuser, self.appointment))
            # User is neither a staff member nor a superuser
            self.assertFalse(has_permission_to_delete_appointment(self.user2, self.appointment))



# email_sender.py
# Path: appointment/email_sender/email_sender.py

from django.core.mail import mail_admins, send_mail
from django.template import loader

from appointment.logger_config import get_logger
from appointment.settings import APP_DEFAULT_FROM_EMAIL, check_q_cluster

logger = get_logger(__name__)

try:
    from django_q.tasks import async_task

    DJANGO_Q_AVAILABLE = True
except ImportError:
    async_task = None
    DJANGO_Q_AVAILABLE = False
    logger.warning("django-q is not installed. Email will be sent synchronously.")


def has_required_email_settings():
    """Check if all required email settings are configured and warn if any are missing."""
    from django.conf import settings as s
    required_settings = [
        'EMAIL_BACKEND', 'EMAIL_HOST', 'EMAIL_PORT',
        'EMAIL_HOST_USER', 'EMAIL_HOST_PASSWORD', 'EMAIL_USE_TLS',
        'EMAIL_USE_LOCALTIME', 'ADMINS',
    ]
    missing_settings = [
        setting_name for setting_name in required_settings if not hasattr(s, setting_name)
    ]

    if missing_settings:
        missing_settings_str = ", ".join(missing_settings)
        logger.warning(f"Warning: The following settings are missing in settings.py: {missing_settings_str}. "
                       "Email functionality will be disabled.")
        return False
    return True


def render_email_template(template_url, context):
    if template_url:
        return loader.render_to_string(template_url, context)
    return ""


def send_email(recipient_list, subject: str, template_url: str = None, context: dict = None, from_email=None,
               message: str = None):
    if not has_required_email_settings():
        return

    from_email = from_email or APP_DEFAULT_FROM_EMAIL
    html_message = render_email_template(template_url, context)

    if get_use_django_q_for_emails() and check_q_cluster() and DJANGO_Q_AVAILABLE:
        # Asynchronously send the email using Django-Q
        async_task(
                "appointment.tasks.send_email_task", recipient_list=recipient_list, subject=subject,
                message=message, html_message=html_message if template_url else None, from_email=from_email
        )
    else:
        # Synchronously send the email
        try:
            send_mail(
                    subject=subject, message=message if not template_url else "",
                    html_message=html_message if template_url else None, from_email=from_email,
                    recipient_list=recipient_list, fail_silently=False,
            )
        except Exception as e:
            logger.error(f"Error sending email: {e}")


def notify_admin(subject: str, template_url: str = None, context: dict = None, message: str = None):
    if not has_required_email_settings():
        return

    html_message = render_email_template(template_url, context)

    if get_use_django_q_for_emails() and check_q_cluster() and DJANGO_Q_AVAILABLE:
        # Enqueue the task to send admin email asynchronously
        async_task('appointment.tasks.notify_admin_task', subject=subject, message=message, html_message=html_message)
    else:
        # Synchronously send admin email
        try:
            mail_admins(
                    subject=subject, message=message if not template_url else "",
                    html_message=html_message if template_url else None
            )
        except Exception as e:
            logger.error(f"Error sending email to admin: {e}")


def get_use_django_q_for_emails():
    """Get the value of the USE_DJANGO_Q_FOR_EMAILS setting."""
    try:
        from django.conf import settings
        return getattr(settings, 'USE_DJANGO_Q_FOR_EMAILS', False)
    except AttributeError:
        logger.error("Error accessing USE_DJANGO_Q_FOR_EMAILS. Defaulting to False.")
        return False



from datetime import timedelta

from django.test import TestCase

from appointment.models import Appointment, AppointmentRequest, Service, StaffMember
from appointment.tests.mixins.base_mixin import (
    AppointmentMixin, AppointmentRequestMixin, AppointmentRescheduleHistoryMixin, ServiceMixin, StaffMemberMixin,
    UserMixin
)
from appointment.utils.db_helpers import get_user_model


class BaseTest(TestCase, UserMixin, StaffMemberMixin, ServiceMixin, AppointmentRequestMixin,
               AppointmentMixin, AppointmentRescheduleHistoryMixin):
    service1 = None
    service2 = None
    staff_member1 = None
    staff_member2 = None
    users = None

    USER_SPECS = {
        'staff1': {"first_name": "Daniel", "last_name": "Jackson", "email": "daniel.jackson@django-appointment.com",
                   "username": "daniel.jackson"},
        'staff2': {"first_name": "Samantha", "last_name": "Carter", "email": "samantha.carter@django-appointment.com",
                   "username": "samantha.carter"},
        'client1': {"first_name": "Georges", "last_name": "Hammond",
                    "email": "georges.s.hammond@django-appointment.com", "username": "georges.hammond"},
        'client2': {"first_name": "Tealc", "last_name": "Kree", "email": "tealc.kree@django-appointment.com",
                    "username": "tealc.kree"},
        'superuser': {"first_name": "Jack", "last_name": "O'Neill", "email": "jack-oneill@django-appointment.com",
                      "username": "jack.o.neill"},
    }

    @classmethod
    def setUpTestData(cls):
        cls.users = {key: cls.create_user_(**details) for key, details in cls.USER_SPECS.items()}
        cls.service1 = cls.create_service_(
            name="Stargate Activation", duration=timedelta(hours=1), price=100000, description="Activate the Stargate")
        cls.service2 = cls.create_service_(
            name="Dial Home Device Repair", duration=timedelta(hours=2), price=200000, description="Repair the DHD")
        # Mapping services to staff members
        cls.staff_member1 = cls.create_staff_member_(user=cls.users['staff1'], service=cls.service1)
        cls.staff_member2 = cls.create_staff_member_(user=cls.users['staff2'], service=cls.service2)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        # Clean up any class-level resources
        cls.clean_all_data()

    @classmethod
    def clean_all_data(cls):
        Appointment.objects.all().delete()
        AppointmentRequest.objects.all().delete()
        StaffMember.objects.all().delete()
        Service.objects.all().delete()
        get_user_model().objects.all().delete()

    def create_appt_request_for_sm1(self, service=None, staff_member=None, **kwargs):
        """Create an appointment request for staff_member1."""
        service = service or self.service1
        staff_member = staff_member or self.staff_member1
        return self.create_appointment_request_(service=service, staff_member=staff_member, **kwargs)

    def create_appt_request_for_sm2(self, service=None, staff_member=None, **kwargs):
        """Create an appointment request for staff_member2."""
        service = service or self.service2
        staff_member = staff_member or self.staff_member2
        return self.create_appointment_request_(service=service, staff_member=staff_member, **kwargs)

    def create_appt_for_sm1(self, appointment_request=None):
        if not appointment_request:
            appointment_request = self.create_appt_request_for_sm1()
        return self.create_appointment_(user=self.users['client1'], appointment_request=appointment_request)

    def create_appt_for_sm2(self, appointment_request=None):
        if not appointment_request:
            appointment_request = self.create_appt_request_for_sm2()
        return self.create_appointment_(user=self.users['client2'], appointment_request=appointment_request)

    def create_appt_reschedule_for_sm1(self, appointment_request=None, reason_for_rescheduling="Gate Malfunction"):
        if not appointment_request:
            appointment_request = self.create_appt_request_for_sm1()
        date_ = appointment_request.date + timedelta(days=7)
        return self.create_reschedule_history_(
            appointment_request=appointment_request,
            date_=date_,
            start_time=appointment_request.start_time,
            end_time=appointment_request.end_time,
            staff_member=appointment_request.staff_member,
            reason_for_rescheduling=reason_for_rescheduling,
        )

    def need_normal_login(self):
        self.client.force_login(self.create_user_())

    def need_staff_login(self):
        self.staff = self.users['staff1']
        self.staff.is_staff = True
        self.staff.save()
        self.client.force_login(self.staff)

    def need_superuser_login(self):
        self.superuser = self.users['superuser']
        self.superuser.is_superuser = True
        self.superuser.save()
        self.client.force_login(self.superuser)

    def clean_staff_member_objects(self, staff=None):
        """Delete all AppointmentRequests and Appointments linked to the StaffMember instance of self.user1."""
        if staff is None:
            staff = self.users['staff1']
        self.clean_appointment_for_user_(staff)
        self.clean_appt_request_for_user_(staff)



# test_session.py
# Path: appointment/tests/utils/test_session.py

from unittest import mock

from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import Client, override_settings
from django.test.client import RequestFactory

from appointment.tests.base.base_test import BaseTest
from appointment.utils.session import (
    get_appointment_data_from_session, handle_email_change, handle_existing_email
)


@override_settings(EMAIL_BACKEND='django.core.mail.backends.locmem.EmailBackend')
class HandleExistingEmailTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.ar = self.create_appt_request_for_sm1()
        self.client = Client()
        self.factory = RequestFactory()

        # Setup request object
        self.request = self.factory.post('/')
        self.hammond = self.users['client1']
        self.request.user = self.hammond

        # Setup session for the request
        session_middleware = SessionMiddleware(lambda req: None)
        session_middleware.process_request(self.request)
        self.request.session.save()

        # Setup messages for the request
        messages_middleware = MessageMiddleware(lambda req: None)
        messages_middleware.process_request(self.request)
        self.request.session.save()

    @mock.patch('appointment.utils.session.get_user_by_email')
    @mock.patch('appointment.utils.session.send_verification_email')
    def test_handle_existing_email(self, mock_send_verification_email, mock_get_user_by_email):
        """Test if an existing email can be handled."""
        client_data = {
            'email': 'georges.s.hammond@django-appointment.com',
            'name': 'georges.hammond',
        }
        appointment_data = {
            'phone': '123456789',
            'want_reminder': True,
            'address': '123, Stargate Command, Cheyenne Mountain, Colorado, USA',
            'additional_info': 'Please bring a Zat gun.'
        }

        response = handle_existing_email(self.request, client_data, appointment_data, self.ar.id, self.ar.id_request)

        # Assert session data
        session = self.request.session
        self.assertEqual(session['email'], client_data['email'])
        self.assertEqual(session['phone'], appointment_data['phone'])
        self.assertTrue(session['want_reminder'])
        self.assertEqual(session['address'], appointment_data['address'])
        self.assertEqual(session['additional_info'], appointment_data['additional_info'])

        # Assert redirect
        self.assertEqual(response.status_code, 302)
        mock_send_verification_email.assert_called_once_with(user=mock_get_user_by_email.return_value,
                                                             email=client_data['email'])
        mock_get_user_by_email.assert_called_once_with(client_data['email'])


class HandleEmailChangeTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.client = Client()
        self.factory = RequestFactory()

        # Setup request object
        self.request = self.factory.post('/')
        self.hammond = self.users['client1']
        self.request.user = self.hammond

        # Setup session for the request
        session_middleware = SessionMiddleware(lambda req: None)
        session_middleware.process_request(self.request)
        self.request.session.save()

    @mock.patch('appointment.utils.session.send_verification_email')
    def test_handle_email_change(self, mock_send_verification_email):
        """Test if an email change can be handled."""
        new_email = 'georges.hammond@django-appointment.com'

        response = handle_email_change(self.request, self.hammond, new_email)

        # Assert session data
        session = self.request.session
        self.assertEqual(session['email'], new_email)
        self.assertEqual(session['old_email'], self.hammond.email)

        # Assert redirect
        self.assertEqual(response.status_code, 302)
        mock_send_verification_email.assert_called_once_with(user=self.hammond, email=new_email)


class GetAppointmentDataFromSessionTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.client = Client()
        self.factory = RequestFactory()

        # Setup request object
        self.request = self.factory.post('/')
        self.hammond = self.users['client1']
        self.request.user = self.hammond

        # Setup session for the request
        session_middleware = SessionMiddleware(lambda req: None)
        session_middleware.process_request(self.request)
        self.request.session.save()

    def test_get_appointment_data_from_session(self):
        """Test if appointment data can be retrieved from the session."""
        # Populate session with test data
        session_data = {
            'phone': '+1234567890',
            'want_reminder': 'on',
            'address': '123 Main St, City, Country',
            'additional_info': 'Some additional info'
        }
        for key, value in session_data.items():
            self.request.session[key] = value
        self.request.session.save()

        # Retrieve data using the function
        appointment_data = get_appointment_data_from_session(self.request)
        self.assertEqual(str(appointment_data['phone']), session_data['phone'])
        self.assertTrue(appointment_data['want_reminder'])
        self.assertEqual(appointment_data['address'], session_data['address'])
        self.assertEqual(appointment_data['additional_info'], session_data['additional_info'])
